#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-11-10
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#

from pathlib import Path
import matplotlib.pyplot as plt
import numpy as np

from robomeshcat import Object, Scene

example_folder = Path(__file__).parent

"All elements are stored in Scene that is responsible for rendering them in the MeshCat"
scene = Scene()

"You can objects into the scene and refer to them through the object instance or by the name"
obj1 = Object.create_sphere(radius=0.1, name='red_sphere', opacity=0.5, color=[1., 0., 0.])
scene.add_object(obj1)

scene.render()
input('Rotate the scene as you wish and press enter to capture the image.')

"Render the image via matplotlib"
img = scene.render_image()

fig, ax = plt.subplots(1, 1, squeeze=True)  # type: plt.Figure, plt.Axes
ax.imshow(img)
plt.show()

"Store the video"
with scene.video_recording(filename='/tmp/video.mp4', fps=30):
    for t in np.linspace(1, 0, 30):
        obj1.pos[2] = t
        scene.render()

    t = np.linspace(0, 2 * np.pi, 60)
    for tt in t:
        scene.camera_pos[0] = np.sin(tt)
        scene.camera_pos[1] = np.cos(tt)
        scene.camera_pos[2] = 1
        scene.render()

scene.reset_camera()
scene.render()



#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-10-13
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#
import itertools
from pathlib import Path
from typing import Union, Dict, Optional, List, Tuple

import numpy as np
import pinocchio as pin

from .object import Object, ArrayWithCallbackOnSetItem


class Robot:
    id_iterator = itertools.count()

    def __init__(self, urdf_path: Union[Path, str], mesh_folder_path: Union[Path, str, List[Path], List[str]] = None,
                 show_collision_models: bool = False, name: str = None, color: Optional[List[float]] = None,
                 opacity: Optional[float] = None, pose=None) -> None:
        """
        Create a robot from URDF using pinocchio loader.
        :param urdf_path: path to the urdf file that contains robot description
        :param mesh_folder_path: either a single path to the directory of meshes or list of paths to meshes directory,
          it's set to directory of urdf_path file by default
        :param show_collision_models: weather to show collision model instead of visual model
        :param name: name of the robot used in meshcat tree
        :param color: optional color that overwrites one from the urdf
        :param opacity: optional opacity that overwrites one from the urdf
        """
        super().__init__()
        self.name = f'robot{next(self.id_iterator)}' if name is None else name
        self._model, self._data, self._geom_model, self._geom_data = self._build_model_from_urdf(
            urdf_path, mesh_folder_path, show_collision_models
        )

        """ Adjustable properties """
        self._pose = ArrayWithCallbackOnSetItem(np.eye(4) if pose is None else pose, cb=self._fk)
        self._q = ArrayWithCallbackOnSetItem(pin.neutral(self._model), cb=self._fk)
        self._color = ArrayWithCallbackOnSetItem(Object._color_from_input(color), cb=self._color_reset_on_set_item)
        self._opacity = opacity
        self._visible = True

        """Set of objects used to visualize the links."""
        self._objects: Dict[str, Object] = {}
        self._init_objects(overwrite_color=color is not None)

    @staticmethod
    def _build_model_from_urdf(urdf_path, mesh_folder_path, show_collision_models) -> Tuple[
        pin.Model, pin.Data, pin.GeometryModel, pin.GeometryData]:
        """Use pinocchio to load models and datas used for the visualizer."""
        if mesh_folder_path is None:
            mesh_folder_path = str(Path(urdf_path).parent)  # by default use the urdf parent directory
        elif isinstance(mesh_folder_path, Path):
            mesh_folder_path = str(mesh_folder_path)
        elif isinstance(mesh_folder_path, list) and any((isinstance(path, Path) for path in mesh_folder_path)):
            mesh_folder_path = [str(path) for path in mesh_folder_path]

        model, col_model, vis_model = pin.buildModelsFromUrdf(str(urdf_path), mesh_folder_path)
        data, col_data, vis_data = pin.createDatas(model, col_model, vis_model)
        geom_model: pin.GeometryModel = col_model if show_collision_models else vis_model
        geom_data: pin.GeometryData = col_data if show_collision_models else vis_data
        return model, data, geom_model, geom_data

    def _fk(self):
        """Compute ForwardKinematics and update the objects poses."""
        pin.forwardKinematics(self._model, self._data, self._q)
        pin.updateGeometryPlacements(self._model, self._data, self._geom_model, self._geom_data)
        base = pin.SE3(self._pose)
        for g, f in zip(self._geom_model.geometryObjects, self._geom_data.oMg):  # type: pin.GeometryObject, pin.SE3
            self._objects[f'{self.name}/{g.name}'].pose = (base * f).homogeneous

    def _init_objects(self, overwrite_color=False):
        """Fill in objects dictionary based on the data from pinocchio"""
        pin.forwardKinematics(self._model, self._data, self._q)
        pin.updateGeometryPlacements(self._model, self._data, self._geom_model, self._geom_data)
        base = pin.SE3(self._pose)
        for g, f in zip(self._geom_model.geometryObjects, self._geom_data.oMg):  # type: pin.GeometryObject, pin.SE3
            kwargs = dict(
                name=f'{self.name}/{g.name}',
                color=g.meshColor[:3] if not overwrite_color else self._color,
                opacity=g.meshColor[3] if self._opacity is None else self._opacity,
                texture=g.meshTexturePath if g.meshTexturePath else None,
                pose=(base * f).homogeneous
            )
            if g.meshPath == 'BOX':
                self._objects[kwargs['name']] = Object.create_cuboid(lengths=2 * g.geometry.halfSide, **kwargs)
            elif g.meshPath == 'SPHERE':
                self._objects[kwargs['name']] = Object.create_sphere(radius=g.geometry.radius, **kwargs)
            elif g.meshPath == 'CYLINDER':
                r, l = g.geometry.radius, 2 * g.geometry.halfLength
                self._objects[kwargs['name']] = Object.create_cylinder(radius=r, length=l, **kwargs)
            else:
                self._objects[kwargs['name']] = Object.create_mesh(path_to_mesh=g.meshPath, scale=g.meshScale, **kwargs)

    """ === Methods for adjusting the base pose of the robot. ==="""

    @property
    def pose(self):
        return self._pose

    @pose.setter
    def pose(self, v):
        self._pose[:, :] = v

    @property
    def pos(self):
        return self._pose[:3, 3]

    @pos.setter
    def pos(self, p):
        self._pose[:3, 3] = p

    @property
    def rot(self):
        return self._pose[:3, :3]

    @rot.setter
    def rot(self, r):
        self._pose[:3, :3] = r

    def __getitem__(self, key):
        jid = self._model.getJointId(key) if isinstance(key, str) else key
        return self._q[jid]

    def __setitem__(self, key, value):
        jid = self._model.getJointId(key) if isinstance(key, str) else key
        self._q[jid] = value

    """=== Control of the object visibility ==="""

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, v):
        self._visible = bool(v)
        for o in self._objects.values():
            o.visible = self._visible

    def hide(self):
        self.visible = False

    def show(self):
        self.visible = True

    """=== Control of the object transparency ==="""

    @property
    def opacity(self):
        return self._opacity

    @opacity.setter
    def opacity(self, v):
        self._opacity = v
        for o in self._objects.values():
            o.opacity = self._opacity

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, v):
        self._color = Object._color_from_input(v)
        self._color_reset_on_set_item()

    def _color_reset_on_set_item(self):
        for o in self._objects.values():
            o.color = self._color



#!/usr/bin/env python
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-10-11
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#

import time
from pathlib import Path
import numpy as np
from robomeshcat import Robot, Scene

"This example shows how to add and control robots online in the visualizer."
"Robots can be loaded from URDF; we will add multiple robots to show the visual/collision models of the robot."

"Example robot data package for the robot meshes and URDF, please install it with"
"conda install -c conda-forge example-robot-data"

"All elements are stored in Scene that is responsible for rendering them in the MeshCat"
scene = Scene()

from example_robot_data.robots_loader import PandaLoader as RobLoader

"Create the first robot and add it to the scene"
rob = Robot(urdf_path=RobLoader().df_path, mesh_folder_path=Path(RobLoader().model_path).parent.parent, name='visual')
scene.add_robot(rob)

"Modify the configuration of the robot"
input('Press enter to change configuration of the robot')
rob[3] = np.deg2rad(-90)  # access by joint id, you can also use scene['visual'][3] = ...
rob['panda_joint5'] = np.deg2rad(90)  # access by joint name, you can also use scene['visual']['panda_joint5'] = ...

input('Press enter to change the pose of the robot')
rob.pos[1] = -1

input('Press enter to change the color of the robot')
rob.color[0] = 1.

input('Press enter to add a new robot that shows collision model of the robot')

col = Robot(urdf_path=RobLoader().df_path, mesh_folder_path=Path(RobLoader().model_path).parent.parent,
            show_collision_models=True)
scene.add_robot(col)
col[:] = rob[:]  # use the same pose for the collision model

input('Press enter to hide visual model and exit')
rob.hide()

time.sleep(1.)  # note that some delay is needed to propagate all the changes



#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-11-30
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#
from pathlib import Path
import numpy as np
import pinocchio as pin
import torch

from robomeshcat import Scene, Human

"This examples show how to use the human model and how to animate its pose and color. "
"We show three case studies: "
"(i) the first case study shows how to manipulate human online, i.e. without animation"
"(ii) this case study shows how to animate pose and color of the human in case of uniform color"
"(iii) this case study shows how use per vertex color of the human (only in online mode, no animation yet!)"

case_study = 2  # chose which case study to visualize
scene = Scene()
"Set smplx_models_path to the directory where are your SMPLX models"
smplx_models_path = str(Path(__file__).parent.joinpath('models').joinpath('smplx'))
human_default_pose = pin.exp6(np.array([0, 0, 0, np.pi / 2, 0., 0.])).homogeneous
human_default_pose[2, 3] = 1.2

if case_study == 0:
    "First let's create the human, arguments are forward to smplx constructor, so you can adjust the human model args"
    human = Human(pose=human_default_pose, color=[1., 0., 0.], model_path=smplx_models_path)
    scene.add_object(human)  # add human to the scene, it will be visualized immediately

    input('Press enter to change the body pose and shape of the human')
    human.smplx_model.body_pose.data += 0.1  # modify the pose
    human.smplx_model.betas.data += 0.1  # modify shape params
    human.smplx_model.expression.data += 0.1  # modify expression param
    human.update_vertices()  # recreate the geometry model to update in the viewer, this is allowed only in online use

    input('Press enter to change the color, opacity, and position of the human')
    human.pos[0] = 1.
    human.color = [0, 1, 0]
    human.opacity = 0.5
    input('Press enter to hide the model and exit.')
    human.hide()

elif case_study == 1:
    human = Human(pose=human_default_pose, color=[1., 0., 0.], model_path=smplx_models_path)

    # You need to create all the animation poses of the human in advance of adding the human to the scene
    # It's called morphologies of the pose
    human.smplx_model.betas.data += 1
    human.add_morph(human.get_vertices())  # the first morph changes the shape

    human.smplx_model.body_pose.data += 0.1
    human.add_morph(human.get_vertices())  # the second morp changes the body pose

    scene.add_object(human)  # add human to the scene, no morphology can be added/modified after this step

    "Let's animate"
    with scene.animation(fps=1):
        human.display_morph(None)  # this will display the human shape that is not affected by morphologies
        scene.render()

        human.display_morph(0)
        scene.render()

        human.display_morph(1)
        scene.render()

        human.color = [0, 0.8, 0]
        scene.render()

        # You can also change the .pos, .rot, .opacity, .visible, in animation
elif case_study == 2:
    # To have per vertex colors, use attribute use_vertex_colors=True
    human = Human(pose=human_default_pose, color=[1., 0., 0.], model_path=smplx_models_path, use_vertex_colors=True)
    scene.add_object(human)

    input('press enter to change colors to random')
    human.update_vertices(vertices_colors=np.random.rand(human.smplx_model.get_num_verts(), 3))

    input('press enter to change colors to blue')
    human.update_vertices(vertices_colors=[[0., 0., 0.75]] * human.smplx_model.get_num_verts())

    input('press enter to display wireframe')
    human._show_wireframe = True
    human.update_vertices(vertices_colors=[[0., 0., 0.]] * human.smplx_model.get_num_verts())

    input('sample new expressions and store them into video, rotate manually to see the face')
    # human.update_vertices(vertices=human.get_vertices(betas=torch.randn([1, 10])))
    human._show_wireframe = False
    human._vertex_colors[:] = 0.6
    with scene.video_recording(filename='/tmp/face_expression.mp4', fps=1):
        for _ in range(10):
            human.update_vertices(vertices=human.get_vertices(expression=torch.randn([1, 10])))
            scene.render()



#!/usr/bin/env python

# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-10-11
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#

from setuptools import setup

if __name__ == '__main__':
    setup()



#!/usr/bin/env python
#
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-11-10
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#

import numpy as np
from pathlib import Path
from example_robot_data.robots_loader import TalosFullLoader as TalosLoader
from example_robot_data.robots_loader import PandaLoader
from robomeshcat import Scene, Robot

scene = Scene()

panda = Robot(urdf_path=PandaLoader().df_path, mesh_folder_path=Path(PandaLoader().model_path).parent.parent)
scene.add_robot(panda)
panda.pos = [1, 0, 0]
panda[3] = -np.pi/2
panda[5] = np.pi/2

talos = Robot(urdf_path=TalosLoader().df_path, mesh_folder_path=Path(TalosLoader().model_path).parent.parent)
scene.add_robot(talos)
talos.pos[2] = 1.075

# with scene.animation(fps=30):
with scene.video_recording(filename='/tmp/teaser.mp4', fps=30):
    scene.camera_pos = [2, 0, 2.5]
    for t in np.linspace(0., 1., 60):
        talos['arm_right_1_joint'] = -t
        talos['arm_left_1_joint'] = t
        panda[0] = t
        panda[2] = t
        scene.render()

    for t in np.linspace(0, 2 * np.pi, 60):
        scene.camera_pos[0] = 2 * np.cos(t)
        scene.camera_pos[1] = 2 * np.sin(t)
        talos['head_1_joint'] = t
        scene.render()




#!/usr/bin/env python
# Copyright (c) CTU -- All Rights Reserved
# Created on: 2022-10-11
#     Author: Vladimir Petrik <vladimir.petrik@cvut.cz>
#

import numpy as np
from pathlib import Path
from pinocchio.utils import rotate

from robomeshcat import Object, Scene

" This example shows how to update properties of the objects online - i.e., changes are apparent in the visualizer "
"immediately as they are set."

"All elements are stored in Scene that is responsible for rendering them in the MeshCat. "
scene = Scene()

"You can add objects into the scene and refer to them through the object instance or by the name"
obj1 = Object.create_sphere(radius=0.1, name='red_sphere', opacity=0.5, color=[1., 0., 0.])
scene.add_object(obj1)


"Let's set object position in z-axis, both options do the same"
input('Press enter to continue: to move sphere up')
obj1.pos[2] = 1.
scene['red_sphere'].pos[2] = 1.


"Other objects could be added to the scene too"
input('Press enter to continue: to add more objects')
scene.add_object(Object.create_cuboid(lengths=0.1, name='box', color=[1, 0, 0]))
scene.add_object(Object.create_cylinder(length=0.6, radius=0.1, name='cylinder'))
scene.add_object(Object.create_mesh(path_to_mesh=Path(__file__).parent.joinpath('spade.obj'), scale=1e-3, name='spade'))

scene['box'].pos[1] = 0.25  # other options: scene['box'].pos = [1, 1, 1], scene['box'].pose = ...
scene['cylinder'].pos[0] = -0.5
scene['cylinder'].rot = rotate('x', np.deg2rad(90))
scene['spade'].pos[1] = 0.75


"You can also adjust color, opacity, or visibility of the object."
input('Press enter to continue: to change the color and opacity')
scene['box'].color = [0, 1, 0]  # other options: scene['box'].color[1] = 1.
scene['box'].opacity = 0.5

input('Press enter to continue: to hide box and finis')
scene['box'].hide()  # other options: scene['box'].visible = False
